var sampleRate = 44100; /* hard-coded in Flash player */

function AudioPlayer(generator, opts) {
	if (!opts) opts = {};
	var latency = opts.latency || 1;
	var checkInterval = latency * 100 /* in ms */
	
	//usingWebAudio = true;

	var audioElement = {};
	if (window["Audio"] != undefined) {
		audioElement = new Audio();
	}
	var webkitAudio = window.AudioContext || window.webkitAudioContext;
	var requestStop = false;
	
	if (webkitAudio) {
		// Uses Webkit Web Audio API if available
		var context = new webkitAudio();
		sampleRate = context.sampleRate;
		
		var channelCount = 2;
		var bufferSize = 4096*4; // Higher for less gitches, lower for less latency
		
		var node = context.createJavaScriptNode(bufferSize, 0, channelCount);
		var gainNode = context.createGainNode();
		gainNode.gain.value = 0.3;
		
		node.onaudioprocess = function(e) { process(e) };

		function process(e) {
			if (generator.finished) {
				node.disconnect();
				return;
			}
			
			var dataLeft = e.outputBuffer.getChannelData(0);
			var dataRight = e.outputBuffer.getChannelData(1);

			var generate = generator.generate(bufferSize);

			for (var i = 0; i < bufferSize; ++i) {
				dataLeft[i] = generate[i*2];
				dataRight[i] = generate[i*2+1];
			}
		}
		
		// start
		//node.connect(context.destination);
		node.connect(gainNode);
		gainNode.connect(context.destination);
		
		return {
			'stop': function() {
				// pause
				gainNode.disconnect();
				node.disconnect();
				requestStop = true;
			},
			'type': 'Webkit Audio',
			'setVolume': function(_vol) {
				gainNode.gain.value = _vol / 100;
				//return true;
			}
		}

	}
	else if (audioElement.mozSetup) {
		audioElement.mozSetup(2, sampleRate); /* channels, sample rate */
		
		var buffer = []; /* data generated but not yet written */
		var minBufferLength = latency * 2 * sampleRate; /* refill buffer when there are only this many elements remaining */
		var bufferFillLength = Math.floor(latency * sampleRate);
		
		function checkBuffer() {
			if (buffer.length) {
				var written = audioElement.mozWriteAudio(buffer);
				buffer = buffer.slice(written);
			}
			if (buffer.length < minBufferLength && !generator.finished) {
				buffer = buffer.concat(generator.generate(bufferFillLength));
			}
			if (!requestStop && (!generator.finished || buffer.length)) {
				setTimeout(checkBuffer, checkInterval);
			}
		}
		checkBuffer();
		
		return {
			'type': 'Firefox Audio',
			'stop': function() {
				requestStop = true;
			},
			'setVolume': function(_vol) {
				audioElement.volume = _vol / 100;
				//return true;
			}
		}
	} else {
		// Fall back to creating flash player
		var RunFlag;
		var flashVer = CFlash.getVerStr();
		if (flashVer < "11.0.0") {
			RunFlag = 0;
			var c = document.createElement('div');
			c.innerHTML = '<embed type="application/x-shockwave-flash" id="da-swf" src="open/script/audio/da.swf" width="8" height="8" allowScriptAccess="always" style="position: fixed; left:-10px;" />';
			document.body.appendChild(c);
			var swf = document.getElementById('da-swf');
		} else {
			RunFlag = 1;
			var BUFF_SIZE = 8192*2;
			var dilayVol = 30;
			/***
			window.getIEAudioData = function () {
				return write(generator.generate(BUFF_SIZE));
			};
			***/
			var flashObj = new CFlash('open/script/audio/da2.swf');
			flashObj.init();
			var swf = flashObj.get();
			if (flashObj.embedResult != CFlash.loadOk) {
				flashObj.createLoadBadHTML();
			}
			//	BUFF_SIZE = swf.getBufferSize();
			//Dbg.putLog("init. checkInterval : "+checkInterval+"ms , BUFF_SIZE : "+BUFF_SIZE);
		}
		
		var minBufferDuration = latency * 1000; /* refill buffer when there are only this many ms remaining */
		var bufferFillLength = latency * sampleRate;
		//putLog("minBufferDuration : "+minBufferDuration);
		//putLog("bufferFillLength : "+bufferFillLength);
		
		function write(data) {
			var out = new Array(data.length);
			for (var i = data.length-1; i != 0; i--) {
				out[i] = Math.floor(data[i]*32768);
			}
			return swf.write(out.join(' '));
		}
		
		function checkBuffer() {
			if (swf.bufferedDuration() < minBufferDuration) {
				write(generator.generate(bufferFillLength));
			};
			if (!requestStop && !generator.finished) setTimeout(checkBuffer, checkInterval);
		}
		
		function write2(data) {
			var out = new Array(data.length);
			/***
			for (var i = data.length-1; i != 0; i--) {
				out[i] = Math.floor(data[i]*32768);
			}
			***/
			for (var i = 0; i < data.length; i++) {
				out[i] = Math.floor(data[i]*32768);
			}
			return swf.write(out.join(' '));
			//return out.join(' ');
		}
		
		function checkBuffer2() {
			if (typeof Dbg !== "undefined")
				Dbg.putAlive(4);
			//putLog("checkBuffer called. "+swf.bufferedDuration()+" < "+minBufferDuration);
			//putLog("checkBuffer called. bufferedDuration : "+swf.bufferedDuration()+" < minBufferDuration : "+minBufferDuration);
			//putLog("checkBuffer(). bufferedDuration : "+swf.bufferedDuration());
			//if (swf.bufferedDuration() <= minBufferDuration) {
			//putLog("checkBuffer2()");
			var duration = swf.bufferedDuration();
			//putLog("checkBuffer2() duration : "+duration);
			//if (swf.bufferedDuration() <= (BUFF_SIZE/2)) {
			if (duration < (BUFF_SIZE)) {
				//var ret = write(generator.generate(bufferFillLength));
				var ret = write2(generator.generate(BUFF_SIZE));
				//putLog("write call. ret : "+ret+" duration : "+duration);
				if (typeof Dbg !== "undefined")
					Dbg.putBuffSize(ret.toString());
			} else {
				//putLog("write pass. bufferedDuration : "+duration);
			}
			if (!requestStop && !generator.finished) setTimeout(checkBuffer2, checkInterval);
		}

		function checkReady() {
			if (swf.write) {
				if (RunFlag) {
					BUFF_SIZE = swf.getBufferSize();
					swf.setVolume(dilayVol / 100);
					flashObj.loaded();
					checkBuffer2();
				} else {
					checkBuffer();
				}
			} else {
				setTimeout(checkReady, 100);
				if (typeof Dbg !== "undefined")
					Dbg.putAlive(3);
			}
		}
		checkReady();
		
		return {
			'stop': function() {
				swf.stop();
				requestStop = true;
			},
			'bufferedDuration': function() {
				return swf.bufferedDuration();
			},
			'type': 'Flash Audio',
			'setVolume': function(_vol) {
				if (flashObj.embedResult == CFlash.loadOk) {
					swf.setVolume(_vol / 100);
				} else {
					dilayVol = _vol;
				}
			}
		}
	}
}
